-module(eRum_app).

-behaviour(application).

-include("eRum.hrl").

-export([
   start/0
   , start/2
   , stop/1
   , doStart/1
   , start_handler/3
   , configure_sink/2
]).



start() ->
   application:ensure_all_started(eRum).

start(_StartType, _StartArgs) ->
   {ok, Pid} = eRum_sup:start_link(),
   SavedHandlers = boot(),
   _ = doStart('__all_extra'),
   _ = doStart('__traces'),
   clean_up_config_checks(),
   {ok, Pid, SavedHandlers}.

boot() ->
   %% Handle the default sink.
   tryStartAsyncer(?RumDefSink, rumUtil:get_env(?RumAppName, asyncThreshold, undefined), rumUtil:get_env(?RumAppName, asyncThresholdWindow, undefined)),

   _ = maybe_install_sink_killer(?RumDefSink, application:get_env(lager, killer_hwm, undefined),
      application:get_env(lager, killer_reinstall_after, undefined)),

   start_handlers(?RumDefSink, application:get_env(lager, handlers, ?DEFAULT_HANDLER_CONF)),

   eRum:update_loglevel_config(?RumDefSink),

   SavedHandlers = start_error_logger_handler(
      application:get_env(lager, error_logger_redirect, true),
      interpret_hwm(application:get_env(lager, error_logger_hwm, 0)),
      application:get_env(lager, error_logger_whitelist, [])
   ),

   SavedHandlers.

doStart('__traces') ->
   _ = rumUtil:trace_filter(none),
   ok = add_configured_traces();

doStart('__all_extra') ->
   configure_extra_sinks(application:get_env(lager, extra_sinks, []));

doStart(?RumDefSink) -> boot();
doStart(Sink) ->
   AllSinksDef = application:get_env(lager, extra_sinks, []),
   boot_sink(Sink, lists:keyfind(Sink, 1, AllSinksDef)).

boot_sink(Sink, {Sink, Def}) ->
   configure_sink(Sink, Def);
boot_sink(Sink, false) ->
   configure_sink(Sink, []).

stop(Handlers) ->
   lists:foreach(fun(Handler) ->
      error_logger:add_report_handler(Handler)
                 end, Handlers),
   rumConfig:cleanup().

tryStartAsyncer(_Sink, undefined, _Window) ->
   ok;
tryStartAsyncer(_Sink, Threshold, _Window) when not is_integer(Threshold) orelse Threshold < 0 ->
   error_logger:error_msg("Invalid value for 'asyncThreshold': ~p~n", [Threshold]),
   throw({error, bad_config});
tryStartAsyncer(Sink, Threshold, undefined) ->
   startThrottle(Sink, Threshold, erlang:trunc(Threshold * 0.2));
tryStartAsyncer(_Sink, Threshold, Window) when not is_integer(Window) orelse Window > Threshold orelse Window < 0 ->
   error_logger:error_msg("Invalid value for 'asyncThresholdWindow': ~p~n", [Window]),
   throw({error, bad_config});
tryStartAsyncer(Sink, Threshold, Window) ->
   startThrottle(Sink, Threshold, Window).

startThrottle(Sink, Threshold, Window) ->
   _ = supervisor:start_child(rumHWatcherSup, [Sink, ?RumBackendThrottle, [Threshold, Window]]),
   ok.

start_handlers(_Sink, undefined) ->
   ok;
start_handlers(_Sink, Handlers) when not is_list(Handlers) ->
   error_logger:error_msg("Invalid value for 'handlers' (must be list): ~p~n", [Handlers]),
   throw({error, bad_config});
start_handlers(Sink, Handlers) ->
   %% handlers failing to start are handled in the handler_watcher
   rumConfig:global_set(handlers,
      rumConfig:global_get(handlers, []) ++
      lists:map(
         fun
            ({Module, Config}) ->
               check_handler_config(Module, Config),
               start_handler(Sink, Module, Config);
            (_) ->
               throw({error, bad_config})
         end,
         expand_handlers(Handlers))),
   ok.

start_handler(Sink, Module, Config) ->
   {ok, Watcher} = supervisor:start_child(lager_handler_watcher_sup, [Sink, Module, Config]),
   {Module, Watcher, Sink}.

check_handler_config({lager_file_backend, F}, Config) when is_list(Config); is_tuple(Config) ->
   Fs =
      case get(?RumPdLogNames) of
         undefined -> ordsets:new();
         X -> X
      end,
   case ordsets:is_element(F, Fs) of
      true ->
         error_logger:error_msg(
            "Cannot have same file (~p) in multiple file backends~n", [F]),
         throw({error, bad_config});
      false ->
         put(?RumPdLogNames,
            ordsets:add_element(F, Fs))
   end,
   ok;
check_handler_config(_Handler, Config) when is_list(Config) orelse is_atom(Config) ->
   ok;
check_handler_config(Handler, _BadConfig) ->
   throw({error, {bad_config, Handler}}).

clean_up_config_checks() ->
   erase(?RumPdLogNames).

interpret_hwm(undefined) ->
   undefined;
interpret_hwm(HWM) when not is_integer(HWM) orelse HWM < 0 ->
   _ = eRum:log(warning, self(), "Invalid error_logger high water mark: ~p, disabling", [HWM]),
   undefined;
interpret_hwm(HWM) ->
   HWM.

maybe_install_sink_killer(_Sink, undefined, _ReinstallTimer) -> ok;
maybe_install_sink_killer(Sink, HWM, undefined) -> maybe_install_sink_killer(Sink, HWM, 5000);
maybe_install_sink_killer(Sink, HWM, ReinstallTimer) when is_integer(HWM) andalso is_integer(ReinstallTimer) andalso HWM >= 0 andalso ReinstallTimer >= 0 ->
   _ = supervisor:start_child(lager_handler_watcher_sup, [Sink, lager_manager_killer,
      [HWM, ReinstallTimer]]);
maybe_install_sink_killer(_Sink, HWM, ReinstallTimer) ->
   error_logger:error_msg("Invalid value for 'killer_hwm': ~p or 'killer_reinstall_after': ~p", [HWM, ReinstallTimer]),
   throw({error, bad_config}).

-spec start_error_logger_handler(boolean(), pos_integer(), list()) -> list().
start_error_logger_handler(false, _HWM, _Whitelist) ->
   [];
start_error_logger_handler(true, HWM, WhiteList) ->
   GlStrategy =
      case application:get_env(lager, error_logger_groupleader_strategy) of
         undefined ->
            handle;
         {ok, GlStrategy0} when
            GlStrategy0 =:= handle;
            GlStrategy0 =:= ignore;
            GlStrategy0 =:= mirror ->
            GlStrategy0;
         {ok, BadGlStrategy} ->
            error_logger:error_msg(
               "Invalid value for 'error_logger_groupleader_strategy': ~p~n",
               [BadGlStrategy]),
            throw({error, bad_config})
      end,

   case whereis(error_logger) of
      undefined ->
         %% On OTP 21 and above, error_logger is deprecated in favor of 'logger'
         %% As a band-aid, boot up error_logger anyway and install it as a logger handler
         %% we can't use error_logger:add_report_handler because we want supervision of the handler
         %% so we have to manually add the logger handler
         %%
         %% Longer term we should be installing a logger handler instead, but this will bridge the gap
         %% for now.
         _ = error_logger:start(),
         _ = logger:add_handler(error_logger, error_logger, #{level => info, filter_default => log}),
         ok = maybe_remove_logger_handler();
      _ ->
         ok
   end,

   %% capture which handlers we removed from error_logger so we can restore them when lager stops
   OldHandlers =
      case supervisor:start_child(lager_handler_watcher_sup, [error_logger, error_logger_lager_h, [HWM, GlStrategy]]) of
         {ok, _} ->
            [begin error_logger:delete_report_handler(X), X end ||
               X <- gen_event:which_handlers(error_logger) -- [error_logger_lager_h | WhiteList]];
         {error, _} ->
            []
      end,
   OldHandlers.

%% On OTP 21.1 and higher we need to remove the `default' handler.
%% But it might not exist, so we will wrap this in a try-catch
%% block
maybe_remove_logger_handler() ->
   try
      ok = logger:remove_handler(default)
   catch
      error:undef -> ok;
      Err:Reason ->
         error_logger:error_msg("calling logger:remove_handler(default) failed: ~p ~p",
            [Err, Reason])
   end.

configure_sink(Sink, SinkDef) ->
   rumConfig:initSink(Sink),
   ChildId = rumUtil:make_internal_sink_name(Sink),
   _ = supervisor:start_child(eRum_sup, {ChildId, {gen_event, start_link, [{local, Sink}]}, permanent, 5000, worker, dynamic}),
   tryStartAsyncer(Sink, proplists:get_value(asyncThreshold, SinkDef), proplists:get_value(asyncThresholdWindow, SinkDef)),
   _ = maybe_install_sink_killer(Sink, proplists:get_value(killer_hwm, SinkDef),
      proplists:get_value(killer_reinstall_after, SinkDef)),
   start_handlers(Sink, proplists:get_value(handlers, SinkDef, [])),
   eRum:update_loglevel_config(Sink).


configure_extra_sinks(Sinks) ->
   lists:foreach(fun({Sink, Proplist}) -> configure_sink(Sink, Proplist) end, Sinks).



expand_handlers([]) ->
   [];
expand_handlers([{lager_file_backend, [{Key, _Value} | _] = Config} | T]) when is_atom(Key) ->
   %% this is definitely a new-style config, no expansion needed
   [maybe_make_handler_id(lager_file_backend, Config) | expand_handlers(T)];
expand_handlers([{lager_file_backend, Configs} | T]) ->
   ?INT_LOG(notice, "Deprecated lager_file_backend config detected, please consider updating it", []),
   [{rumBackendFile:config_to_id(Config), Config} || Config <- Configs] ++
   expand_handlers(T);
expand_handlers([{Mod, Config} | T]) when is_atom(Mod) ->
   [maybe_make_handler_id(Mod, Config) | expand_handlers(T)];
expand_handlers([H | T]) ->
   [H | expand_handlers(T)].

add_configured_traces() ->
   Traces = case application:get_env(lager, traces) of
               undefined ->
                  [];
               {ok, TraceVal} ->
                  TraceVal
            end,

   lists:foreach(fun start_configured_trace/1, Traces),
   ok.

start_configured_trace({Handler, Filter}) ->
   {ok, _} = eRum:trace(Handler, Filter);
start_configured_trace({Handler, Filter, Level}) when is_atom(Level) ->
   {ok, _} = eRum:trace(Handler, Filter, Level).

maybe_make_handler_id(Mod, Config) ->
   %% Allow the backend to generate a gen_event handler id, if it wants to.
   %% We don't use erlang:function_exported here because that requires the module
   %% already be loaded, which is unlikely at this phase of startup. Using code:load
   %% caused undesirable side-effects with generating code-coverage reports.
   try Mod:config_to_id(Config) of
      Id ->
         {Id, Config}
   catch
      error:undef ->
         {Mod, Config}
   end.